/* bitmap.c version 2.1.2 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "bitmap.h"


static unsigned char *write_int4_to_buffer(unsigned char *ptr, int value);
static unsigned char *write_int2_to_buffer(unsigned char *ptr, int value);
static int read_int4_from_buffer(unsigned char *ptr, unsigned char **next);
static int read_int2_from_buffer(unsigned char *ptr, unsigned char **next);
static int write_bitmap_header(FILE *fp, const struct bitmap_header_t *header);
static int write_bitmap_palette(FILE *fp, const struct bitmap_palette_t *palette, int count);
static int check_bitmap_header(struct bitmap_header_t *header);
static int calc_length(int width, int bit_count);
static int read_bitmap_palette(FILE *fp, struct bitmap_palette_t *palette, int count);


static unsigned char *write_int4_to_buffer(unsigned char *ptr, int value)
{
	*ptr++ = (char)(value & 0xFF);
	value >>= 8;
	*ptr++ = (char)(value & 0xFF);
	value >>= 8;
	*ptr++ = (char)(value & 0xFF);
	value >>= 8;
	*ptr++ = (char)(value & 0xFF);

	return ptr;
}

static unsigned char *write_int2_to_buffer(unsigned char *ptr, int value)
{
	*ptr++ = (char)(value & 0xFF);
	value >>= 8;
	*ptr++ = (char)(value & 0xFF);

	return ptr;
}

static int read_int4_from_buffer(unsigned char *ptr, unsigned char **next)
{
	int value = 0;

	value = *ptr++;
	value |= ((int)*ptr++) << 8;
	value |= ((int)*ptr++) << 16;
	value |= ((int)*ptr++) << 24;

	if (next != NULL) {
		*next = ptr;
	}

	return value;
}

static int read_int2_from_buffer(unsigned char *ptr, unsigned char **next)
{
	int value = 0;

	value = *ptr++;
	value |= ((int)*ptr++) << 8;

	if (next != NULL) {
		*next = ptr;
	}

	return value;
}

int create_bitmap_header(struct bitmap_header_t *header, unsigned int width, unsigned int height, unsigned short bit_count)
{
	header->reserved1 = 0;
	header->reserved2 = 0;
	header->header_size = 40;
	header->width = width;
	header->height = height;
	header->planes = 1;
	header->bit_count = bit_count;
	header->compression = 0;
	header->h_resolution = 0;
	header->v_resolution = 0;
	header->used_colors = 0;
	header->important_colors = 0;
	switch (bit_count) {
	case 1:
		header->offset_bitmap = 54 + (4 * 2);
		header->data_length = (((width + 7) / 8 + 3) & ~3) * height;
		header->file_length = 54 + header->data_length + (4 * 2);
		break;
	case 4:
		header->offset_bitmap = 54 + (4 * 16);
		header->data_length = (((width + 1) / 2 + 3) & ~3) * height;
		header->file_length = 54 + header->data_length + (4 * 16);
		break;
	case 8:
		header->offset_bitmap = 54 + (4 * 256);
		header->data_length = ((width + 3) & ~3) * height;
		header->file_length = 54 + header->data_length + (4 * 256);
		break;
	case 16:
		header->offset_bitmap = 54;
		header->data_length = ((width * 2 + 3) & ~3) * height;
		header->file_length = 54 + header->data_length;
		break;
	case 24:
		header->offset_bitmap = 54;
		header->data_length = ((width * 3 + 3) & ~3) * height;
		header->file_length = 54 + header->data_length;
		break;
	case 32:
		header->offset_bitmap = 54;
		header->data_length = width * 4 * height;
		header->file_length = 54 + header->data_length;
		break;
	default:
		return -1;
	}
	return 0;
}

static int write_bitmap_header(FILE *fp, const struct bitmap_header_t *header)
{
	unsigned char buf[54];
	unsigned char *ptr = buf;

	*ptr++ = 'B';
	*ptr++ = 'M';
	ptr = write_int4_to_buffer(ptr, header->file_length);
	ptr = write_int2_to_buffer(ptr, header->reserved1);
	ptr = write_int2_to_buffer(ptr, header->reserved2);
	ptr = write_int4_to_buffer(ptr, header->offset_bitmap);
	ptr = write_int4_to_buffer(ptr, header->header_size);
	ptr = write_int4_to_buffer(ptr, header->width);
	ptr = write_int4_to_buffer(ptr, header->height);
	ptr = write_int2_to_buffer(ptr, header->planes);
	ptr = write_int2_to_buffer(ptr, header->bit_count);
	ptr = write_int4_to_buffer(ptr, header->compression);
	ptr = write_int4_to_buffer(ptr, header->data_length);
	ptr = write_int4_to_buffer(ptr, header->h_resolution);
	ptr = write_int4_to_buffer(ptr, header->v_resolution);
	ptr = write_int4_to_buffer(ptr, header->used_colors);
	ptr = write_int4_to_buffer(ptr, header->important_colors);

	if (fwrite(buf, sizeof(buf), 1, fp) != 1) {
		return -1;
	}

	return 0;
}

static int write_bitmap_palette(FILE *fp, const struct bitmap_palette_t *palette, int count)
{
	unsigned char buf[4 * 256];
	unsigned char *ptr = buf;
	int i;

	if (count < 0 || 256 < count) {
		return -1;
	}

	for (i = 0; i < count; i++, palette++) {
		*ptr++ = palette->blue;
		*ptr++ = palette->green;
		*ptr++ = palette->red;
		*ptr++ = palette->alpha;
	}

	if (fwrite(buf, count * 4, 1, fp) != 1) {
		return -1;
	}

	return 0;
}

int write_bitmap(FILE *fp, const struct bitmap_header_t *header, const struct bitmap_palette_t *palette, const unsigned char *data)
{
	if (header->bit_count != 1 &&
		header->bit_count != 4 &&
		header->bit_count != 8 &&
		header->bit_count != 16 &&
		header->bit_count != 24 &&
		header->bit_count != 32) {
		return -1;
	}

	if (write_bitmap_header(fp, header) < 0) {
		printf("error: write_bitmap() - can't write bitmap header\n");
		return -1;
	}

	if (header->bit_count <= 8) {
		int i = 1 << header->bit_count;
		if (palette == NULL) {
			unsigned int dummy = 0;
			while (i--) {
				if (fwrite(&dummy, 4, 1, fp) != 1) {	
					printf("error: write_bitmap() - can't write palette\n");
					return -1;
				}
			}
		} else {
			if (write_bitmap_palette(fp, palette, i) < 0) {
				printf("error: write_bitmap() - can't write palette\n");
				return -1;
			}
		}
	}

	if (header->bit_count == 1) {
		int y = header->height;
		int x;
		int xmax;
		int b;
		int val;
		int pad;
		int p;
		const unsigned char *d;
		const unsigned char mask[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

		pad = ((header->width + 7) / 8) & 3;
		if (pad != 0) {
			pad = 4 - pad;
		}

		while (y--) {
			xmax = header->width / 8;
			d = data + header->width * y;
			for (x = 0; x < xmax; x++) {
				val = 0;
				for (b = 0; b < 8; b++) {
					if (*d++) {
						val |= mask[b];
					}
				}
				fputc(val, fp);
			}
			if (header->width & 7) {
				val = 0;
				for (b = 0; b < (header->width & 7); b++) {
					if (*d++) {
						val |= mask[b];
					}
				}
				fputc(val, fp);
			}
			p = pad;
			while (p--) {
				fputc(0, fp);
			}
		}
	} else if (header->bit_count == 4) {
		int y = header->height;
		int x;
		int xmax;
		int pad;
		int p;
		const unsigned char *d;

		pad = ((header->width + 1) / 2) & 3;
		if (pad != 0) {
			pad = 4 - pad;
		}

		while (y--) {
			xmax = header->width >> 1;
			d = data + header->width * y;
			for (x = 0; x < xmax; x++, d += 2) {
				fputc(((*d & 0x0F) << 4) | (*(d + 1) & 0x0F), fp);
			}
			if (header->width & 1) {
				fputc((*d & 0x0F) << 4, fp);
			}
			p = pad;
			while (p--) {
				fputc(0, fp);
			}
		}
	} else {
		int y = header->height;
		int unit = header->bit_count / 8;
		int pad;
		int p;

		if (header->bit_count == 32) {
			pad = 0;
		} else {
			pad = (header->width * unit) & 3;
			if (pad != 0) {
				pad = 4 - pad;
			}
		}

		while (y--) {
			fwrite(data + header->width * y * unit, unit, header->width, fp);
			p = pad;
			while (p--) {
				fputc(0, fp);
			}
		}
	}

	return 0;
}

static int calc_length(int width, int bit_count)
{
	if (bit_count == 1) {
		return ((width + 7) / 8 + 3) & ~3;
	}

	if (bit_count == 4) {
		return ((width + 1) / 2 + 3) & ~3;
	}

	if (bit_count == 8 ||
		bit_count == 16 ||
		bit_count == 24) {
		return (width * (bit_count / 8) + 3) & ~3;
	}

	if (bit_count == 32) {
		return width * 4;
	}

	return -1;
}

static int check_bitmap_header(struct bitmap_header_t *header)
{
	int warning_count = 0;
	int palette_size;
	int offset_bitmap;
	int data_length;
	int file_length;

	if (header->header_size != 40) {
		printf("error: check_bitmap_header() - not supported header type - support only Windows V3\n");
		return -1;
	}

	if (header->planes != 1) {
		return -1;
	}

	if (header->compression != 0) {
		printf("error: check_bitmap_header() - not support compression\n");
		return -1;
	}

	if (header->bit_count == 1 || header->bit_count == 4 || header->bit_count == 8) {
		palette_size = 4 << header->bit_count;
	} else if (header->bit_count == 16 || header->bit_count == 24 || header->bit_count == 32) {
		palette_size = 0;
	} else {
		printf("error: check_bitmap_header() - unknown bit count\n");
		return -1;
	}

	offset_bitmap = 54 + palette_size;
	data_length = calc_length(header->width, header->bit_count) * header->height;
	file_length = offset_bitmap + data_length;

	if (header->offset_bitmap != offset_bitmap) {
		printf("error: check_bitmap_header() - offset size 0x%X != 0x%X\n", header->offset_bitmap, offset_bitmap);
		return -1;
	}

	if (header->data_length == 0) {
		header->data_length = data_length;
		printf("warning: check_bitmap_header() - data length is 0 ... corrected: 0x%X\n", data_length);
		warning_count++;
	} if (header->data_length != data_length) {
		printf("error: check_bitmap_header() - data length 0x%X != 0x%X\n", header->data_length, data_length);
		return -1;
	}

	if (header->file_length > file_length) {
		printf("warning: check_bitmap_header() - file length 0x%X > 0x%X\n", header->file_length, file_length);
		warning_count++;
	} else if (header->file_length < file_length) {
		printf("error: check_bitmap_header() - file length 0x%X < 0x%X\n", header->file_length, file_length);
		return -1;
	}

	return warning_count;
}

int read_bitmap_header(FILE *fp, struct bitmap_header_t *header)
{
	unsigned char buf[54];
	unsigned char *ptr = buf;

	if (fread(buf, sizeof(buf), 1, fp) != 1) {
		return -1;
	}

	if (ptr[0] != 'B' || ptr[1] != 'M') {
		return -1;
	}
	ptr += 2;

	header->file_length		 = read_int4_from_buffer(ptr, &ptr);
	header->reserved1		 = read_int2_from_buffer(ptr, &ptr);
	header->reserved2		 = read_int2_from_buffer(ptr, &ptr);
	header->offset_bitmap	 = read_int4_from_buffer(ptr, &ptr);
	header->header_size		 = read_int4_from_buffer(ptr, &ptr);
	header->width			 = read_int4_from_buffer(ptr, &ptr);
	header->height			 = read_int4_from_buffer(ptr, &ptr);
	header->planes			 = read_int2_from_buffer(ptr, &ptr);
	header->bit_count		 = read_int2_from_buffer(ptr, &ptr);
	header->compression		 = read_int4_from_buffer(ptr, &ptr);
	header->data_length		 = read_int4_from_buffer(ptr, &ptr);
	header->h_resolution	 = read_int4_from_buffer(ptr, &ptr);
	header->v_resolution	 = read_int4_from_buffer(ptr, &ptr);
	header->used_colors		 = read_int4_from_buffer(ptr, &ptr);
	header->important_colors = read_int4_from_buffer(ptr, &ptr);

	return check_bitmap_header(header);
}

static int read_bitmap_palette(FILE *fp, struct bitmap_palette_t *palette, int count)
{
	unsigned char buf[4 * 256];
	unsigned char *ptr = buf;
	int i;

	if (count < 0 || 256 < count) {
		return -1;
	}

	if (fread(buf, count * 4, 1, fp) != 1) {
		return -1;
	}

	for (i = 0; i < count; i++, palette++) {
		palette->blue  = *ptr++;
		palette->green = *ptr++;
		palette->red   = *ptr++;
		palette->alpha = *ptr++;
	}

	return 0;
}

int read_bitmap(FILE *fp, struct bitmap_header_t *header, struct bitmap_palette_t **palette, unsigned char **data, int header_already_read)
{
	unsigned char *work;
	size_t length;
	unsigned int y;

	if (data == NULL) {
		return -3;
	}

	if (! header_already_read) {
		if (read_bitmap_header(fp, header) < 0) {
			printf("error: read_bitmap() - can't read bitmap header\n");
			return -1;
		}
	}

	if (*data == NULL) {
		*data = (unsigned char *)malloc(header->width * header->height * ((header->bit_count + 7) / 8));
		if (*data == NULL) {
			printf("error: memory exhaust!\n");
			return -2;
		}
	}

	if (header->bit_count == 1 || header->bit_count == 4 || header->bit_count == 8) {
		int palette_count = 1 << header->bit_count;
		if (palette == NULL) {
			fseek(fp, 4 * palette_count, SEEK_CUR);
		} else {
			if (*palette == NULL) {
				*palette = (struct bitmap_palette_t *)malloc(sizeof(struct bitmap_palette_t) * palette_count);
				if (*palette == NULL) {
					printf("error: memory exhaust!\n");
					return -2;
				}
			}

			if (read_bitmap_palette(fp, *palette, palette_count) < 0) {
				printf("error: read_bitmap() - can't read palette\n");
				return -1;
			}
		}
	}

	length = calc_length(header->width, header->bit_count);
	if ((work = (unsigned char *)malloc(length)) == NULL) {
		printf("error: memory exhaust!\n");
		return -2;
	}

	y = header->height;
	while (0 < y--) {
		unsigned char *w;

		if (fread(work, length, 1, fp) != 1) {
			free(work);
			printf("error: read_bitmap() - can't read bitmap data\n");
			return -1;
		}
		w = work;

		if (header->bit_count == 1) {
			unsigned char *d = (*data) + header->width * y;
			unsigned int x = header->width / 8;
			int b;
			const unsigned char mask[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
			while (x--) {
				int val = *w++;
				for (b = 0; b < 8; b++) {
					*d++ = (val & mask[b]) ? 1 : 0;
				}
			}
			if (header->width & 7) {
				int val = *w++;
				for (b = 0; b < (header->width & 7); b++) {
					*d++ = (val & mask[b]) ? 1 : 0;
				}
			}
		} else if (header->bit_count == 4) {
			unsigned char *d = (*data) + header->width * y;
			unsigned int x = header->width / 2;
			while (x--) {
				*d++ = *w >> 4;
				*d++ = *w++ & 0x0F;
			}
			if (header->width & 1) {
				*d = *w >> 4;
			}
		} else {
			const int line_size = (header->bit_count / 8) * header->width;
			unsigned char *d = (*data) + line_size * y;
			unsigned int x = line_size;
			while (x--) {
				*d++ = *w++;
			}
		}
	}
	free(work);

	return 0;
}

int find_palette_best_match(int red, int green, int blue, const struct bitmap_palette_t *palette, int count)
{
	int match = 0;
	int match_score = 0x30000;
	int i;

	for (i = 0; i < count; i++) {
		int b = (int)palette[i].blue  - blue;
		int g = (int)palette[i].green - green;
		int r = (int)palette[i].red   - red;
		int score = b * b + g * g + r * r;

		if (score == 0) {
			return i;
		}

		if (score < match_score) {
			match = i;
			match_score = score;
		}
	}

	return match;
}

int convert_bit_count(int width, int height, int src_bit_count, const struct bitmap_palette_t *src_palette, const unsigned char *src_data, int des_bit_count, const struct bitmap_palette_t *des_palette, unsigned char **des_data)
{
	int src_palette_count = 0;
	int des_palette_count = 0;
	int des_data_size;
	const unsigned char *src;
	unsigned char *des;
	int y;
	int x;
	const struct bitmap_palette_t mono_palette[2] = {
		RGB(0x00, 0x00, 0x00),
		RGB(0xFF, 0xFF, 0xFF)
	};

	if (src_data == NULL || des_data == NULL) {
		return -1;
	}

	des_data_size = width * height * ((des_bit_count + 7) / 8);
	if (*des_data == NULL) {
		*des_data = (unsigned char *)malloc(des_data_size);
		if (*des_data == NULL) {
			printf("error: memory exhaust!\n");
			return -2;
		}
	}
	memset(*des_data, 0, des_data_size);

	src = src_data;
	des = *des_data;

	if (src_bit_count == 1) {
		if (src_palette == NULL) {
			src_palette = mono_palette;
		}
		src_palette_count = 2;
	} else if (src_bit_count == 4 || src_bit_count == 8) {
		if (src_palette == NULL) {
			return -1;
		}
		src_palette_count = src_bit_count == 4 ? 16 : 256;
	} else if (src_bit_count != 16 && src_bit_count != 24 && src_bit_count != 32) {
		return -1;
	}

	if (des_bit_count == 1) {
		if (des_palette == NULL) {
			des_palette = mono_palette;
		}
		des_palette_count = 2;
	} else if (des_bit_count == 4 || des_bit_count == 8) {
		if (des_palette == NULL) {
			return -1;
		}
		des_palette_count = des_bit_count == 4 ? 16 : 256;
	} else if (des_bit_count != 16 && des_bit_count != 24 && des_bit_count != 32) {
		return -1;
	}

	if (src_palette_count != 0) {
		if (des_palette_count != 0) {
			/* 1 or 4 or 8 -> 1 or 4 or 8 */
			unsigned char trans_palette[256];
			int i;

			for (i = 0; i < src_palette_count; i++) {
				trans_palette[i] = find_palette_best_match(src_palette[i].red,
														   src_palette[i].green,
														   src_palette[i].blue,
														   des_palette,
														   des_palette_count);
			}

			for (y = 0; y < height; y++) {
				for (x = 0; x < width; x++) {
					*des++ = trans_palette[*src++];
				}
			}
		} else if (des_bit_count == 16) {
			/* 1 or 4 or 8 -> 16 */
			for (y = 0; y < height; y++) {
				for (x = 0; x < width; x++) {
					int c = *src++;
					int b = src_palette[c].blue >> 3;
					int g = src_palette[c].green >> 3;
					int r = src_palette[c].red >> 3;
					int color = (r << 10) | (g << 5) | (b);
					*des++ = color & 0xFF;
					*des++ = color >> 8;
				}
			}
		} else {
			/* 1 or 4 or 8 -> 24 or 32 */
			for (y = 0; y < height; y++) {
				for (x = 0; x < width; x++) {
					int c = *src++;
					*des++ = src_palette[c].blue;
					*des++ = src_palette[c].green;
					*des++ = src_palette[c].red;
					if (des_bit_count == 32) {
						*des++ = 0;
					}
				}
			}
		}
	} else if (des_palette_count != 0) {
		if (src_bit_count == 16) {
			/* 16 -> 1 or 4 or 8 */
			for (y = 0; y < height; y++) {
				for (x = 0; x < width; x++) {
					int color = *src | (*(src + 1) << 8);
					int b =  color        & 0x1F;
					int g = (color >>  5) & 0x1F;
					int r = (color >> 10) & 0x1F;
					b = (b << 3) | (b >> 2);
					g = (g << 3) | (g >> 2);
					r = (r << 3) | (r >> 2);
					src += 2;
					*des++ = find_palette_best_match(r, g, b, des_palette, des_palette_count);
				}
			}
		} else {
			/* 24 or 32 -> 1 or 4 or 8 */
			for (y = 0; y < height; y++) {
				for (x = 0; x < width; x++) {
					int b = *src++;
					int g = *src++;
					int r = *src++;
					if (src_bit_count == 32) {
						src++;
					}
					*des++ = find_palette_best_match(r, g, b, des_palette, des_palette_count);
				}
			}
		}
	} else if (src_bit_count == des_bit_count) {
		/* copy */
		int count = width * height * (src_bit_count / 8);
		while (count--) {
			*des++ = *src++;
		}
	} else if (src_bit_count == 16) {
		/* 16 -> 24 or 32 */
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				int color = *src | (*(src + 1) << 8);
				int b =  color        & 0x1F;
				int g = (color >>  5) & 0x1F;
				int r = (color >> 10) & 0x1F;
				src += 2;
				*des++ = (b << 3) | (b >> 2);
				*des++ = (g << 3) | (g >> 2);
				*des++ = (r << 3) | (r >> 2);
				if (des_bit_count == 32) {
					*des++ = 0;
				}
			}
		}
	} else if (des_bit_count == 16) {
		/* 24 or 32 -> 16 */
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				int b = *src++ >> 3;
				int g = *src++ >> 3;
				int r = *src++ >> 3;
				int color = (r << 10) | (g << 5) | (b);
				*des++ = color & 0xFF;
				*des++ = color >> 8;
				if (src_bit_count == 32) {
					src++;
				}
			}
		}
	} else {
		/* 24 -> 32 or 32 -> 24 */
		for (y = 0; y < height; y++) {
			for (x = 0; x < width; x++) {
				*des++ = *src++;
				*des++ = *src++;
				*des++ = *src++;
				if (des_bit_count == 32) {
					*des++ = 0;
				} else {
					src++;
				}
			}
		}
	}

	return 0;
}

/* EOF */
