#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>

#include "bitmap.h"

#include "u5shape_info.h"


#define SHAPE_WIDTH			(16)
#define SHAPE_HEIGHT		(16)

#define NUM_ROWS_TILES		(16)
#define NUM_ROWS_NPC		(16)
#define NUM_ROWS			(NUM_ROWS_TILES + NUM_ROWS_NPC)
#define NUM_COLS			(16)

#define ROW_OFFSET_TILES	(0)
#define ROW_OFFSET_NPC 		(NUM_ROWS_TILES)

#define GRASS_COLOR_EGA		(2)
#define GROUND_COLOR_EGA	(4)

#define GRASS_COLOR_CGA		(1)
#define GROUND_COLOR_CGA	(2)

static const struct bitmap_palette_t shape_palette_ega[16] = {
	RGB(0x00, 0x00, 0x00),		/*  0 black         */
	RGB(0x00, 0x00, 0xAA),		/*  1 blue          */
	RGB(0x00, 0xAA, 0x00),		/*  2 green         */
	RGB(0x00, 0xAA, 0xAA),		/*  3 cyan          */
	RGB(0xAA, 0x00, 0x00),		/*  4 red           */
	RGB(0xAA, 0x00, 0xAA),		/*  5 magenta       */
	RGB(0xAA, 0x55, 0x00),		/*  6 brown         */
	RGB(0xAA, 0xAA, 0xAA),		/*  7 gray          */
	RGB(0x55, 0x55, 0x55),		/*  8 dark gray     */
	RGB(0x55, 0x55, 0xFF),		/*  9 light blue    */
	RGB(0x55, 0xFF, 0x55),		/* 10 light green   */
	RGB(0x55, 0xFF, 0xFF),		/* 11 light cyan    */
	RGB(0xFF, 0x55, 0x55),		/* 12 light red     */
	RGB(0xFF, 0x55, 0xFF),		/* 13 light magenta */
	RGB(0xFF, 0xFF, 0x55),		/* 14 light yellow  */
	RGB(0xFF, 0xFF, 0xFF)		/* 15 white         */
};

static const unsigned char mirror_room_color_ega[16] = {
	0,		/* black                        */
	5,		/* blue          -> magenta     */
	4,		/* green         -> red         */
	4,		/* cyan          -> red         */
	2,		/* red           -> green       */
	1,		/* magenta       -> blue        */
	2,		/* brown         -> green       */
	7,		/* gray                         */
	8,		/* dark gray                    */
	12,		/* light blue    -> light red   */
	12,		/* light green   -> light red   */
	12,		/* light cyan    -> light red   */
	10,		/* light red     -> light green */
	9,		/* light magenta -> light blue  */
	14,		/* light yellow                 */
	15		/* white                        */
};

static const struct bitmap_palette_t shape_palette_cga[16] = {
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0xAA, 0xAA),
	RGB(0xAA, 0x00, 0xAA),
	RGB(0xAA, 0xAA, 0xAA),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00),
	RGB(0x00, 0x00, 0x00)
};

static const unsigned char mirror_room_color_cga[16] = {
	0,		/* black              */
	2,		/* cyan    -> magenta */
	1,		/* magenta -> cyan    */
	3,		/* white              */
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0,
	0
};

#define TILES_SIZE_EGA			(0x10000)
#define TILES_SIZE_CGA			(0x8000)
static unsigned char tiles_data[TILES_SIZE_EGA];
static unsigned char vram[SHAPE_WIDTH * SHAPE_HEIGHT * NUM_ROWS * NUM_COLS];



static void usage(void);
static int read_tiles(const char *filename, unsigned char *data, int data_size);
static int get_shape_offset(int row, int col, int y, int x, int pixel_per_byte);
static int get_shape_data(const unsigned char *shape_data, int row, int col, int y, int x, int pixel_per_byte);
static int remake_shape_data(const unsigned char *shape_data, int row, int col, int y, int x, int pixel_per_byte, int base_shape, int mask_shape);
static unsigned char make_reverse_mask(unsigned char mask, int pixel_per_byte);
static int draw_shape(const unsigned char *shape_data, int row, int col, int pixel_per_byte, int remake_water);
static void change_grass_color(unsigned char grass_color, unsigned char ground_color);
static void change_mirror_room_color(const unsigned char *color);
static int write_shape_bmp(const char *name, int bit_count, const struct bitmap_palette_t *palette, const unsigned char *vram);
static int confirm(const char *msg);



static void usage(void)
{
	printf("usage: u5ega_shapes [options] TILES OUTPUT.BMP\n");
	printf("option: --cga               read cga version tiles\n");
	printf("        --dungeon           change color of grass shapes\n");
	printf("        --mirror-room       use alternate colors for mirror room\n");
	printf("        --no-remake-water   do not remake water shapes\n");
	printf("        --help              print usage\n");
	printf("\n");
	printf("support decoded TILES.16 and TILES.4 only\n");
}

int main(int argc, char **argv)
{
	int num_cmd_params = 2;
	const char *tiles_file = NULL;
	const char *bmp_file = NULL;

	int opt_cga = 0;
	int opt_dungeon = 0;
	int opt_mirror_room = 0;
	int opt_remake_water = 1;
	int row;
	int col;

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "--cga") == 0) {
				opt_cga = 1;
			} else if (strcmp(*argv, "--dungeon") == 0) {
				opt_dungeon = 1;
			} else if (strcmp(*argv, "--mirror-room") == 0) {
				opt_mirror_room = 1;
			} else if (strcmp(*argv, "--no-remake-water") == 0) {
				opt_remake_water = 0;
			} else {
				printf("unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("unneeded parameter '%s'\n", *argv);
				return 1;
			}
			if (num_cmd_params == 2) {
				tiles_file = *argv;
			} else {
				bmp_file = *argv;
			}
			num_cmd_params--;
		}
	}

	if (tiles_file == NULL) {
		printf("no tiles file\n");
		return 1;
	}

	if (bmp_file == NULL) {
		printf("no output bitmap file\n");
		return 1;
	}

	if (read_tiles(tiles_file, tiles_data, (opt_cga ? TILES_SIZE_CGA : TILES_SIZE_EGA)) < 0) {
		printf("read tiles file error '%s'\n", tiles_file);
		return 1;
	}

	memset(vram, 0, sizeof(vram));
	for (row = 0; row < NUM_ROWS; row++) {
		for (col = 0; col < NUM_COLS; col++) {
			draw_shape(tiles_data, row, col, (opt_cga ? 4 : 2), opt_remake_water);
		}
	}

	if (opt_dungeon) {
		if (opt_cga) {
			change_grass_color(GRASS_COLOR_CGA, GROUND_COLOR_CGA);
		} else {
			change_grass_color(GRASS_COLOR_EGA, GROUND_COLOR_EGA);
		}
	}

	if (opt_mirror_room) {
		if (opt_cga) {
			change_mirror_room_color(mirror_room_color_cga);
		} else {
			change_mirror_room_color(mirror_room_color_ega);
		}
	}

	if (write_shape_bmp(bmp_file, 4, (opt_cga ? shape_palette_cga : shape_palette_ega), vram) < 0) {
		printf("write bitmap file error '%s'\n", bmp_file);
		return 1;
	}

	return 0;
}

static int read_tiles(const char *filename, unsigned char *data, int data_size)
{
	int result = -1;
	FILE *fp = NULL;

	if ((fp = fopen(filename, "rb")) == NULL) {
		printf("open error '%s'\n", filename);
		goto ERROR_EXIT;
	}

	if (fread(data, data_size, 1, fp) != 1) {
		printf("read error '%s'\n", filename);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

static int get_shape_offset(int row, int col, int y, int x, int pixel_per_byte)
{
	return (((row * NUM_COLS + col) * SHAPE_HEIGHT + y) * SHAPE_WIDTH + x) / pixel_per_byte;
}

static int get_shape_data(const unsigned char *shape_data, int row, int col, int y, int x, int pixel_per_byte)
{
	return *(shape_data + get_shape_offset(row, col, y, x, pixel_per_byte));
}

static int remake_shape_data(const unsigned char *shape_data, int row, int col, int y, int x, int pixel_per_byte, int base_shape, int mask_shape)
{
	unsigned char base_data;
	unsigned char data;
	int reverse = 0;
	unsigned char mask;
	unsigned char reverse_mask;

	base_data = get_shape_data(shape_data, base_shape >> 4, base_shape & 0xF, y, x, pixel_per_byte);
	data = get_shape_data(shape_data, row, col, y, x, pixel_per_byte);

	if (mask_shape < 0) {
		reverse = 1;
		mask_shape = -mask_shape;
	}

	mask = get_shape_data(shape_data, mask_shape >> 4, mask_shape & 0xF, y, x, pixel_per_byte);
	if (reverse) {
		mask = make_reverse_mask(mask, pixel_per_byte);
		reverse_mask = ~mask;
	} else {
		reverse_mask = make_reverse_mask(mask, pixel_per_byte);
		mask = ~reverse_mask;
	}

	return (base_data & mask) | (data & reverse_mask);
}

static unsigned char make_reverse_mask(unsigned char mask, int pixel_per_byte)
{
	const int bit_count = 8 / pixel_per_byte;
	const unsigned char bit_mask = (unsigned char)((1 << bit_count) - 1);
	unsigned char reverse_mask = 0;
	int i;

	for (i = 0; i < 8; i += bit_count) {
		if ((mask & (bit_mask << i)) == 0) {
			reverse_mask |= (bit_mask << i);
		}
	}

	return reverse_mask;
}

static int draw_shape(const unsigned char *shape_data, int row, int col, int pixel_per_byte, int remake_water)
{
	const int bit_count = 8 / pixel_per_byte;
	const unsigned char bit_mask = (unsigned char)((1 << bit_count) - 1);
	int y;
	int x;
	int mask_shape = (remake_water ? get_water_mask(row, col) : 0);

	for (y = 0; y < SHAPE_HEIGHT; y++) {
		unsigned char *des = vram + (row * SHAPE_HEIGHT + y) * SHAPE_WIDTH * NUM_COLS + col * SHAPE_WIDTH;

		for (x = 0; x < SHAPE_WIDTH; x += pixel_per_byte) {
			unsigned char data;
			int i;

			if (mask_shape) {
				data = remake_shape_data(shape_data, row, col, y, x, pixel_per_byte, 0x03, mask_shape);
			} else {
				data = get_shape_data(shape_data, row, col, y, x, pixel_per_byte);
			}

			i = 8;
			while (0 < i) {
				i -= bit_count;
				*des++ = ((data >> i) & bit_mask);
			}
		}
	}

	return 0;
}

static void change_grass_color(unsigned char grass_color, unsigned char ground_color)
{
	int row;
	int col;
	int x;
	int y;

	for (row = 0; row < NUM_ROWS; row++) {
		for (col = 0; col < NUM_COLS; col++) {
			if (grass_to_ground(row, col)) {
				for (y = 0; y < SHAPE_HEIGHT; y++) {
					unsigned char *ptr = vram + (row * SHAPE_HEIGHT + y) * SHAPE_WIDTH * NUM_COLS + col * SHAPE_WIDTH;

					for (x = 0; x < SHAPE_WIDTH; x++, ptr++) {
						if (*ptr == grass_color) {
							*ptr = ground_color;
						}
					}
				}
			}
		}
	}
}

static void change_mirror_room_color(const unsigned char *color)
{
	int count = sizeof(vram);
	unsigned char *ptr = vram;

	while (0 < count--) {
		*ptr = color[*ptr];
		ptr++;
	}
}

static int write_shape_bmp(const char *name, int bit_count, const struct bitmap_palette_t *palette, const unsigned char *vram)
{
	int result = -1;
	FILE *fp = NULL;
	struct bitmap_header_t header;
	struct stat st;

	if (stat(name, &st) == 0) {
		printf("already exists '%s'\n", name);
		if (confirm("overwrite ?") <= 0) {
			goto ERROR_EXIT;
		}
	}

	if ((fp = fopen(name, "wb")) == NULL) {
		printf("open error '%s'\n", name);
		goto ERROR_EXIT;
	}

	create_bitmap_header(&header, SHAPE_WIDTH * NUM_COLS, SHAPE_HEIGHT * NUM_ROWS, bit_count);
	if (write_bitmap(fp, &header, palette, vram) < 0) {
		printf("write error '%s'\n", name);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

static int confirm(const char *msg)
{
	printf("%s (Y or N): ", msg);
	fflush(stdout);

	while (1) {
		int c = getchar();
		if (c == 'y' || c == 'Y') {
			return 1;
		}
		if (c == 'n' || c == 'N') {
			return 0;
		}
	}

	return -1;
}

/* EOF */
