#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>

#include "bitmap.h"
#include "dosfd_img.h"


#define SHAPE_WIDTH			(32)
#define SHAPE_HEIGHT		(32)

#define NUM_ROWS_TILES		(16)
#define NUM_ROWS_NPC		(16)
#define NUM_ROWS			(NUM_ROWS_TILES + NUM_ROWS_NPC)
#define NUM_COLS			(16)

#define ROW_OFFSET_TILES	(0)
#define ROW_OFFSET_NPC 		(NUM_ROWS_TILES)


static const struct bitmap_palette_t shape_palette[16] = {
	RGB(0x00,0x00,0x00),		/* 0 black */
	RGB(0x00,0x00,0xFF),		/* 1 blue */
	RGB(0xFF,0x00,0x00),		/* 2 red */
	RGB(0xFF,0x00,0xFF),		/* 3 magenta */
	RGB(0x00,0xFF,0x00),		/* 4 green */
	RGB(0x00,0xFF,0xFF),		/* 5 cyan */
	RGB(0xFF,0xFF,0x00),		/* 6 yellow */
	RGB(0xFF,0xFF,0xFF),		/* 7 white */
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00),
	RGB(0x00,0x00,0x00)
};

static const struct bitmap_palette_t mask_palette[2] = {
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF)
};

static unsigned char tiles_data[0x18000];
static unsigned char npc_data[0x20000];
static unsigned char vram[SHAPE_WIDTH * SHAPE_HEIGHT * NUM_ROWS * NUM_COLS];
static unsigned char mask_vram[SHAPE_WIDTH * SHAPE_HEIGHT * NUM_ROWS * NUM_COLS];


static void usage(void);
static char *create_write_name(const char *name);
static void merge_shapes_and_mask(unsigned char *vram, const unsigned char *mask_vram, int size);
static int get_vram_offset(int row, int col, int y);
static int change_shape(int row_offset, int row, int col, unsigned char *data, int bit_count, const unsigned char *vram);
static int read_shapes_bmp(const char *name, int *rows, int bit_count, const struct bitmap_palette_t *palette, unsigned char *vram);
static int confirm(const char *msg);


static void usage(void)
{
#ifdef U5_PORT_X68K
	printf("usage: u5x68kchg_shapes [options] SHAPES.BMP [MASK.BMP]\n");
#else
	printf("usage: u5pc98chg_shapes [options] SHAPES.BMP [MASK.BMP]\n");
#endif
	printf("option: --disk DISK-IMG     disk image name of Britannia Disk\n");
#ifdef U5_PORT_X68K
	printf("                            read 'u5-britannia.xdf' if not specified\n");
#else
	printf("                            read 'U5_Britania.fdi' if not specified\n");
#endif
	printf("        --write DISK-IMG    disk image name of modified Britannia Disk\n");
	printf("        --only-tiles        change landform shapes (TILES.CH) only\n");
	printf("        --only-npc          change npc shapes (NPC.CH) only\n");
	printf("        --help              print usage\n");
	printf("\n");
	printf("remove all mask if mask bitmap is not specified\n");
#ifdef U5_PORT_X68K
	printf("supported disk image format is *.xdf only\n");
#else
	printf("supported disk image format is *.fdi only\n");
#endif
	printf("write modified disk image to filename.NEW.ext if not specified\n");
}

int main(int argc, char **argv)
{
	int num_cmd_params = 2;
	const char *bmp_file = NULL;
	const char *mask_file = NULL;

	const char *tiles_file = "TILES.CH";
	const char *npc_file = "NPC.CH";
#ifdef U5_PORT_X68K
	const char *disk_file = "u5-britannia.xdf";
	const int disk_offset = 0;
#else
	const char *disk_file = "U5_Britania.fdi";
	const int disk_offset = FDI_OFFSET;
#endif
	const char *write_file = NULL;
	int opt_change_tiles = 1;
	int opt_change_npc = 1;
	struct dosfd_image_t image;
	int row;
	int col;
	int bmp_rows;
	int mask_rows;

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "--disk") == 0) {
				if (*++argv == NULL) {
					printf("no disk image (britannia disk)\n");
					return 1;
				}
				disk_file = *argv;
			} else if (strcmp(*argv, "--write") == 0) {
				if (*++argv == NULL) {
					printf("no disk image (output)\n");
					return 1;
				}
				write_file = *argv;
			} else if (strcmp(*argv, "--only-tiles") == 0) {
				opt_change_tiles = 1;
				opt_change_npc = 0;
			} else if (strcmp(*argv, "--only-npc") == 0) {
				opt_change_tiles = 0;
				opt_change_npc = 1;
			} else {
				printf("unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("unneeded parameter '%s'\n\n", *argv);
				return 1;
			} else {
				if (num_cmd_params == 2) {
					bmp_file = *argv;
				} else {
					mask_file = *argv;
				}
				num_cmd_params--;
			}
		}
	}

	if (bmp_file == NULL) {
		printf("no shapes bitmap file\n");
		return 1;
	}

	if (write_file == NULL) {
		write_file = create_write_name(disk_file);
		if (write_file == NULL) {
			printf("can't create filename\n");
			return 1;
		}
	}

	memset(vram, 0, sizeof(vram));
	memset(mask_vram, 0, sizeof(mask_vram));

	/* read shapes bitmap */
	if (read_shapes_bmp(bmp_file, &bmp_rows, 4, shape_palette, vram) < 0) {
		printf("read shapes file error '%s'\n", bmp_file);
		return 1;
	}
	if ((opt_change_npc && bmp_rows < NUM_ROWS) ||
		(!opt_change_npc && bmp_rows < NUM_ROWS_TILES)) {
		printf("shapes bitmap size error '%s'\n", bmp_file);
		return 1;
	}

	/* read mask bitmap */
	if (mask_file != NULL) {
		if (read_shapes_bmp(mask_file, &mask_rows, 1, mask_palette, mask_vram) < 0) {
			printf("read mask file error '%s'\n", mask_file);
			return 1;
		}
		if (bmp_rows != mask_rows) {
			printf("shapes and mask size mismatch\n");
			return 1;
		}
	}

	/* merge shapes and mask */
	merge_shapes_and_mask(vram, mask_vram, sizeof(vram));

	/* read disk image */
	if (read_dosfd_image(&image, disk_file, disk_offset) < 0) {
		printf("read disk image error '%s'\n", disk_file);
		return 1;
	}
	if (read_dosfd_image_file(&image, tiles_file, tiles_data, sizeof(tiles_data), 1) < 0) {
		printf("read disk image error '%s/%s'\n", disk_file, tiles_file);
		return 1;
	}
	if (read_dosfd_image_file(&image, npc_file, npc_data, sizeof(npc_data), 1) < 0) {
		printf("read disk image error '%s/%s'\n", disk_file, npc_file);
		return 1;
	}

	/* change shapes */
	if (opt_change_tiles) {
		for (row = 0; row < NUM_ROWS_TILES; row++) {
			for (col = 0; col < NUM_COLS; col++) {
				change_shape(ROW_OFFSET_TILES, row, col, tiles_data, 3, vram);
			}
		}
	}

	if (opt_change_npc) {
		for(row = 0; row < NUM_ROWS_NPC; row++) {
			for (col = 0; col < NUM_COLS; col++) {
				change_shape(ROW_OFFSET_NPC, row, col, npc_data, 4, vram);
			}
		}
	}

	/* write modified disk image */
	{
		struct stat st;

		if (stat(write_file, &st) == 0) {
			printf("already exists '%s'\n", write_file);
			if (confirm("overwrite ?") <= 0) {
				return 1;
			}
		}

		if (opt_change_tiles &&
			write_dosfd_image_file(&image, tiles_file, tiles_data, sizeof(tiles_data), 1) < 0) {
			printf("write disk image error '%s/%s'\n", write_file, tiles_file);
			return 1;
		}

		if (opt_change_npc &&
			write_dosfd_image_file(&image, npc_file, npc_data, sizeof(npc_data), 1) < 0) {
			printf("write disk image error '%s/%s'\n", write_file, npc_file);
			return 1;
		}

		if (write_dosfd_image(&image, write_file) < 0) {
			printf("write disk image error '%s'\n", write_file);
			return 1;
		}
	}

	free_dosfd_image(&image, 0);

	return 0;
}

static char *create_write_name(const char *name)
{
	int baselen;
	const char *ext;
	int len = strlen(name) + 5;
	char *new_name = (char *)malloc(len);
	if (new_name == NULL) {
		printf("memory exhaust!\n");
		return NULL;
	}

	if ((ext = strrchr(name, '.')) == NULL) {
		baselen = strlen(name);
		ext = "";
	} else {
		baselen = ext - name;
	}

	strncpy(new_name, name, baselen);
	strcpy(new_name + baselen, ".NEW");
	strcpy(new_name + baselen + 4, ext);

	return new_name;
}

static void merge_shapes_and_mask(unsigned char *vram, const unsigned char *mask_vram, int size)
{
	while (0 < size--) {
		if (*mask_vram) {
			*vram |= 0x08;
		}
		mask_vram++;
		vram++;
	}
}

static int get_vram_offset(int row, int col, int y)
{
	return (row * SHAPE_HEIGHT + y) * SHAPE_WIDTH * NUM_COLS + col * SHAPE_WIDTH;
}

static int change_shape(int row_offset, int row, int col, unsigned char *data, int bit_count, const unsigned char *vram)
{
	int plane;
	int y;
	int x;
	int pal;

	data += (row * NUM_COLS + col) * (SHAPE_WIDTH / 8) * SHAPE_HEIGHT * bit_count;

	for (plane = 0; plane < bit_count; plane++) {
		pal = (1 << plane);

		for (y = 0; y < SHAPE_HEIGHT; y++) {
			const unsigned char *src = vram + get_vram_offset(row_offset + row, col, y);

			for (x = 0; x < SHAPE_WIDTH; x += 8) {
				unsigned char val = 0;
				int bit;

				for (bit = 0; bit < 8; bit++, src++) {
					const int mask[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };

					if (*src & pal) {
						val |= mask[bit];
					}
				}

				*data++ = val;
			}
		}
	}

	return 0;
}

static int read_shapes_bmp(const char *name, int *rows, int bit_count, const struct bitmap_palette_t *palette, unsigned char *vram)
{
	int result = -1;
	FILE *fp = NULL;
	struct bitmap_header_t header;
	struct bitmap_palette_t *work_palette = NULL;
	unsigned char *work_data = NULL;

	if ((fp = fopen(name, "rb")) == NULL) {
		printf("open error '%s'\n", name);
		goto ERROR_EXIT;
	}

	if (read_bitmap(fp, &header, &work_palette, &work_data, 0) < 0) {
		printf("read error '%s'\n", name);
		goto ERROR_EXIT;
	}

	fclose(fp);
	fp = NULL;

	if (header.width != SHAPE_WIDTH * NUM_COLS ||
		header.height < SHAPE_HEIGHT * NUM_ROWS_TILES || SHAPE_HEIGHT * NUM_ROWS < header.height) {
		printf("bitmap size error '%s' %d x %d\n", name, header.width, header.height);
		goto ERROR_EXIT;
	}

	*rows = header.height / SHAPE_HEIGHT;

	if (convert_bit_count(header.width, header.height,
						  header.bit_count, work_palette, work_data,
						  bit_count, palette, &vram) < 0) {
		printf("convert bitmap error '%s'\n", name);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}
	if (work_palette != NULL) {
		free(work_palette);
	}
	if (work_data != NULL) {
		free(work_data);
	}

	return result;
}

static int confirm(const char *msg)
{
	printf("%s (Y or N): ", msg);
	fflush(stdout);

	while (1) {
		int c = getchar();
		if (c == 'y' || c == 'Y') {
			return 1;
		}
		if (c == 'n' || c == 'N') {
			return 0;
		}
	}

	return -1;
}

/* EOF */
