#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/stat.h>
#include <errno.h>

#include "bitmap.h"
#include "dosfd_img.h"

#include "u5shape_info.h"


#define SHAPE_WIDTH			(32)
#define SHAPE_HEIGHT		(32)

#define NUM_ROWS_TILES		(16)
#define NUM_ROWS_NPC		(16)
#define NUM_ROWS			(NUM_ROWS_TILES + NUM_ROWS_NPC)
#define NUM_COLS			(16)

#define ROW_OFFSET_TILES	(0)
#define ROW_OFFSET_NPC 		(NUM_ROWS_TILES)


static const struct bitmap_palette_t shape_palette[16] = {
	RGB(0x00,0x00,0x00),		/* 0 black */
	RGB(0x00,0x00,0xFF),		/* 1 blue */
	RGB(0xFF,0x00,0x00),		/* 2 red */
	RGB(0xFF,0x00,0xFF),		/* 3 magenta */
	RGB(0x00,0xFF,0x00),		/* 4 green */
	RGB(0x00,0xFF,0xFF),		/* 5 cyan */
	RGB(0xFF,0xFF,0x00),		/* 6 yellow */
	RGB(0xFF,0xFF,0xFF),		/* 7 white */
	RGBA(0x00,0x00,0x00,0xFF),
	RGBA(0x00,0x00,0xFF,0xFF),
	RGBA(0xFF,0x00,0x00,0xFF),
	RGBA(0xFF,0x00,0xFF,0xFF),
	RGBA(0x00,0xFF,0x00,0xFF),
	RGBA(0x00,0xFF,0xFF,0xFF),
	RGBA(0xFF,0xFF,0x00,0xFF),
	RGBA(0xFF,0xFF,0xFF,0xFF)
};

static const unsigned char mirror_room_color[16] = {
	0,		/* black            */
	5,		/* blue    -> cyan  */
	4,		/* red     -> green */
	4,		/* magenta -> green */
	2,		/* green   -> red   */
	1,		/* cyan    -> blue  */
	2,		/* yello   -> red   */
	7,		/* white            */
	8,
	13,
	12,
	12,
	10,
	9,
	10,
	15,
};

static const struct bitmap_palette_t mask_palette[16] = {
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF)
};

static const struct bitmap_palette_t mono_palette[2] = {
	RGBA(0x00, 0x00, 0x00, 0x00),
	RGBA(0xFF, 0xFF, 0xFF, 0xFF)
};

static unsigned char tiles_data[0x18000];
static unsigned char npc_data[0x20000];
static unsigned char vram[SHAPE_WIDTH * SHAPE_HEIGHT * NUM_ROWS * NUM_COLS];
static unsigned char mask_vram[SHAPE_WIDTH * SHAPE_HEIGHT * NUM_ROWS * NUM_COLS];


static void usage(void);
static int get_vram_offset(int row, int col, int y);
static int draw_shape(int row_offset, int row, int col, const unsigned char *data, int bit_count, int swap_red_green);
static void remake_water_shape(void);
static void change_mirror_room_color(void);
static int write_shapes_bmp(const char *name, int bit_count, const struct bitmap_palette_t *palette, const unsigned char *vram);
static int confirm(const char *msg);


static void usage(void)
{
#ifdef U5_PORT_X68K
	printf("usage: u5x68kshapes [options] OUTPUT.BMP [MASK.BMP]\n");
#else
	printf("usage: u5pc98shapes [options] OUTPUT.BMP [MASK.BMP]\n");
#endif
	printf("option: --disk DISK-IMG     disk image name of Britannia Disk\n");
#ifdef U5_PORT_X68K
	printf("                            read 'u5-britannia.xdf' if not specified\n");
#else
	printf("                            read 'U5_Britania.fdi' if not specified\n");
#endif
	printf("        --dungeon           change color of grass shapes\n");
	printf("        --mirror-room       use alternate colors for mirror room\n");
	printf("        --no-remake-water   do not remake water shapes\n");
	printf("        --help              print usage\n");
	printf("\n");
#ifdef U5_PORT_X68K
	printf("supported disk image format is *.xdf only\n");
#else
	printf("supported disk image format is *.fdi only\n");
#endif
}

int main(int argc, char **argv)
{
	int num_cmd_params = 2;
	const char *tiles_file = "TILES.CH";
	const char *npc_file = "NPC.CH";
#ifdef U5_PORT_X68K
	const char *disk_file = "u5-britannia.xdf";
	const int disk_offset = 0;
#else
	const char *disk_file = "U5_Britania.fdi";
	const int disk_offset = FDI_OFFSET;
#endif
	const char *bmp_file = NULL;
	const char *mask_file = NULL;
	int opt_dungeon = 0;
	int opt_mirror_room = 0;
	int opt_remake_water = 1;
	struct dosfd_image_t image;
	int row;
	int col;

	while (*++argv != NULL) {
		if (**argv == '-') {
			if (strcmp(*argv, "--help") == 0) {
				usage();
				return 1;
			} else if (strcmp(*argv, "--disk") == 0) {
				if (*++argv == NULL) {
					printf("no disk image (britannia disk)\n");
					return 1;
				}
				disk_file = *argv;
			} else if (strcmp(*argv, "--dungeon") == 0) {
				opt_dungeon = 1;
			} else if (strcmp(*argv, "--mirror-room") == 0) {
				opt_mirror_room = 1;
			} else if (strcmp(*argv, "--no-remake-water") == 0) {
				opt_remake_water = 0;
			} else {
				printf("unknown option '%s'\n", *argv);
				return 1;
			}
		} else {
			if (num_cmd_params <= 0) {
				printf("unneeded parameter '%s'\n\n", *argv);
				return 1;
			} else {
				if (num_cmd_params == 2) {
					bmp_file = *argv;
				} else {
					mask_file = *argv;
				}
				num_cmd_params--;
			}
		}
	}

	if (bmp_file == NULL) {
		printf("no output bitmap file\n");
		return 1;
	}

	if (read_dosfd_image(&image, disk_file, disk_offset) < 0) {
		printf("read disk image error '%s'\n", disk_file);
		return 1;
	}
	if (read_dosfd_image_file(&image, tiles_file, tiles_data, sizeof(tiles_data), 1) < 0) {
		printf("read disk image error '%s/%s'\n", disk_file, tiles_file);
		return 1;
	}
	if (read_dosfd_image_file(&image, npc_file, npc_data, sizeof(npc_data), 1) < 0) {
		printf("read disk image error '%s/%s'\n", disk_file, npc_file);
		return 1;
	}
	free_dosfd_image(&image, 0);

	memset(vram, 0, sizeof(vram));
	for (row = 0; row < NUM_ROWS_TILES; row++) {
		for (col = 0; col < NUM_COLS; col++) {
			draw_shape(ROW_OFFSET_TILES, row, col, tiles_data, 3, opt_dungeon);
		}
	}

	for(row = 0; row < NUM_ROWS_NPC; row++) {
		for (col = 0; col < NUM_COLS; col++) {
			draw_shape(ROW_OFFSET_NPC, row, col, npc_data, 4, 0);
		}
	}

	if (opt_remake_water) {
		remake_water_shape();
	}

	if (opt_mirror_room) {
		change_mirror_room_color();
	}

	if (write_shapes_bmp(bmp_file, 4, shape_palette, vram) < 0) {
		printf("write bitmap file error '%s'\n", bmp_file);
		return 1;
	}

	if (mask_file != NULL) {
		unsigned char *ptr = mask_vram;
		memset(mask_vram, 0, sizeof(mask_vram));
		if (convert_bit_count(SHAPE_WIDTH * NUM_COLS, SHAPE_HEIGHT * NUM_ROWS,
							  4, mask_palette, vram,
							  1, mono_palette, &ptr) < 0) {
			printf("convert error\n");
			return 1;
		}

		if (write_shapes_bmp(mask_file, 1, mono_palette, mask_vram) < 0) {
			printf("write bitmap file error '%s'\n", mask_file);
			return 1;
		}
	}

	return 0;
}

static int get_vram_offset(int row, int col, int y)
{
	return (row * SHAPE_HEIGHT + y) * SHAPE_WIDTH * NUM_COLS + col * SHAPE_WIDTH;
}

static int draw_shape(int row_offset, int row, int col, const unsigned char *data, int bit_count, int swap_red_green)
{
	int plane;
	int y;
	int x;
	int pal;

	data += (row * NUM_COLS + col) * (SHAPE_WIDTH / 8) * SHAPE_HEIGHT * bit_count;

	for (plane = 0; plane < bit_count; plane++) {
		pal = (1 << plane);
		if (swap_red_green && grass_to_ground(row, col)) {
			if (pal == 0x2) {
				pal = 0x4;
			} else if (pal == 0x4) {
				pal = 0x2;
			}
		}

		for (y = 0; y < SHAPE_HEIGHT; y++) {
			unsigned char *des = vram + get_vram_offset(row_offset + row, col, y);

			for (x = 0; x < SHAPE_WIDTH; x += 8) {
				unsigned char val = *data++;
				int bit;

				for (bit = 0; bit < 8; bit++, des++) {
					const int mask[8] = { 0x80, 0x40, 0x20, 0x10, 0x08, 0x04, 0x02, 0x01 };
					if (val & mask[bit]) {
						*des |= pal;
					}
				}
			}
		}
	}

	return 0;
}

static void remake_water_shape(void)
{
	int row;
	int col;

	for (row = 0; row < NUM_ROWS_TILES; row++) {
		for (col = 0; col < NUM_COLS; col++) {
			int mask_shape = get_water_mask(row, col);
			int y;
			int x;

			if (mask_shape) {
				int reverse = 0;
				if (mask_shape < 0) {
					reverse = 1;
					mask_shape = -mask_shape;
				}

				if (0xD0 <= mask_shape && mask_shape <= 0xD3) {
					if (reverse) {
						if (mask_shape == 0xD0) {
							mask_shape = 0xD2;
						} else if (mask_shape == 0xD1) {
							mask_shape = 0xD3;
						} else if (mask_shape == 0xD2) {
							mask_shape = 0xD0;
						} else if (mask_shape == 0xD3) {
							mask_shape = 0xD1;
						}
					}
					for (y = 0; y < SHAPE_HEIGHT; y++) {
						const unsigned char *water = vram + get_vram_offset(0, 3, y);
						unsigned char *data = vram + get_vram_offset(row, col, y);
						for (x = 0; x < SHAPE_WIDTH; x++) {
							int mask = 0;
							if (mask_shape == 0xD0) {
								mask = ((x + y) <= 31);
							} else if (mask_shape == 0xD1) {
								mask = ((y - x) <= 0);
							} else if (mask_shape == 0xD2) {
								mask = (31 <= (x + y));
							} else if (mask_shape == 0xD3) {
								mask = (0 <= (y - x));
							}

							if (mask) {
								*data = *water;
							}

							water++;
							data++;
						}
					}
				} else {
					for (y = 0; y < SHAPE_HEIGHT; y++) {
						const unsigned char *mask = vram + get_vram_offset(mask_shape >> 4, mask_shape & 0xF, y);
						const unsigned char *water = vram + get_vram_offset(0, 3, y);
						unsigned char *data = vram + get_vram_offset(row, col, y);

						for (x = 0; x < SHAPE_WIDTH; x++) {
							if ((!reverse && *mask) || (reverse && !*mask)) {
								*data = *water;
							}

							mask++;
							water++;
							data++;
						}
					}
				}
			}
		}
	}
}

static void change_mirror_room_color(void)
{
	int count = sizeof(vram);
	unsigned char *ptr = vram;

	while (0 < count--) {
		*ptr = mirror_room_color[*ptr];
		ptr++;
	}
}

static int write_shapes_bmp(const char *name, int bit_count, const struct bitmap_palette_t *palette, const unsigned char *vram)
{
	int result = -1;
	FILE *fp = NULL;
	struct bitmap_header_t header;
	struct stat st;

	if (stat(name, &st) == 0) {
		printf("already exists '%s'\n", name);
		if (confirm("overwrite ?") <= 0) {
			goto ERROR_EXIT;
		}
	}

	if ((fp = fopen(name, "wb")) == NULL) {
		printf("open error '%s'\n", name);
		goto ERROR_EXIT;
	}

	create_bitmap_header(&header, SHAPE_WIDTH * NUM_COLS, SHAPE_HEIGHT * NUM_ROWS, bit_count);
	if (write_bitmap(fp, &header, palette, vram) < 0) {
		printf("write error '%s'\n", name);
		goto ERROR_EXIT;
	}

	result = 0;

ERROR_EXIT:
	if (fp != NULL) {
		fclose(fp);
	}

	return result;
}

static int confirm(const char *msg)
{
	printf("%s (Y or N): ", msg);
	fflush(stdout);

	while (1) {
		int c = getchar();
		if (c == 'y' || c == 'Y') {
			return 1;
		}
		if (c == 'n' || c == 'N') {
			return 0;
		}
	}

	return -1;
}

/* EOF */
